local Lib = MissionMinderPreLoad;
local DEBUG = Lib.debug;
Lib.callbacks = Lib.callbacks or LibStub("CallbackHandler-1.0"):New(Lib)

Lib.tracker = {};
local module = Lib.tracker;

local L = LibStub("AceLocale-3.0"):GetLocale("MissionMinder");
local COMPLETED = L["completed"]
local L_GARRISON_REPORT_SUMMARY_MISSIONS = GARRISON_MISSIONS
local L_GARRISON_REPORT_SUMMARY_MISSIONS_DESCRIPTION = L["Mission summary for your characters"]
local L_GARRISON_REPORT_SUMMARY_CACHE = GARRISON_CACHE
local L_GARRISON_REPORT_SUMMARY_CACHE_DESCRIPTION = L["Garrison caches that are nearly full"]
local L_GARRISON_REPORT_SUMMARY_BUIDLINGS = GARRISON_LANDING_STATUS_BUILDING
local L_GARRISON_REPORT_SUMMARY_BUIDLINGS_DESCRIPTION = L["Building construction status"]
local L_GARRISON_REPORT_SUMMARY_WORKORDERS = CAPACITANCE_WORK_ORDERS
local L_GARRSION_REPORT_SUMMARY_WORKORDERS_DESCRIPTION = L["Work orders nearing completion"]


local SECONDS = 1
local MINUTES = 60 * SECONDS
local HOURS = 60 * MINUTES
local DAYS = 24 * HOURS


local MAX_GARRISON_CACHE = 500
local GARRISON_CACHE_RATE_IN_SECONDS = 10 * MINUTES
local GARRISON_CACHE_WARN_AMOUNT = MAX_GARRISON_CACHE - (1 * DAYS / GARRISON_CACHE_RATE_IN_SECONDS)

local GARRISON_RESOURCES_TEXTURE=select(3, GetCurrencyInfo(824))

local CURRENT_PLAYER = GetUnitName("player") .. "-" .. GetRealmName()

local TIMES = { DAYS, HOURS, MINUTES}
local TIMESTRS = {INT_SPELL_DURATION_DAYS, INT_SPELL_DURATION_HOURS, INT_SPELL_DURATION_MIN }
function module.secondsToString(seconds)
	local str = ""

	if  ( seconds < MINUTES ) then
		str = str .. string.format(INT_SPELL_DURATION_SEC, seconds);
	else
		local reportOn=2
		for i=1, #TIMES do
			if ( seconds >= TIMES[i] ) then
				if(reportOn <= 0) then
					break;
				end

				if ( str:len() > 0 ) then
					str = str .. " ";
				end

				str = str .. string.format(TIMESTRS[i], seconds / TIMES[i])
				seconds = seconds % TIMES[i]
				reportOn = reportOn - 1
			end
		end
	end

	return str
end

local function createInlineTexture(texture)
	return texture and ("|T" .. texture .. ":0|t") or nil;
end


function module.estimateGarrisonCache(player)
	local now = time()
	local lastLooted = Lib.Globals.Characters[player].GarrisonCache

	if ( lastLooted ) then
		return min(MAX_GARRISON_CACHE, floor((now - lastLooted) / GARRISON_CACHE_RATE_IN_SECONDS))
	else
		return nil
	end
end

local function orderGarrisonCache(cache1, cache2)
	if ( cache1.amount ~= cache2.amount ) then
		return cache1.amount > cache2.amount
	else
		return strcmputf8i(cache1.name, cache1.name) < 0
	end
end

local function orderBuildings(building1, building2)
	if ( building1.player ~= building2.player ) then
		return strcmputf8i(building1.player, building2.player) < 0
	elseif ( building1.isComplete ~= building2.isComplete ) then
		return building1.isComplete
	elseif ( building1.endDate ~= building2.endDate ) then
		return building1.endDate < building2.endDate
	else
		return strcmputf8i(building1.name, building2.name) < 0
	end
end

local function orderWorkOrders(workorder1, workorder2)
	local pending1 = workorder1.total - workorder1.ready
	local pending2 = workorder2.total - workorder2.ready

	if ( workorder1.player ~= workorder2.player ) then
		return strcmputf8i(workorder1.player, workorder2.player) < 0
	elseif ( pending1 ~= pending2 ) then
		return pending1 < pending2
	else
		return strcmputf8i(workorder1.name, workorder2.name) < 0
	end
end

local function orderTrackedMissions(track1, track2)
	if ( track1.completed ~= track2.completed ) then
		return track1.completed > track2.completed
	elseif ( track1.lowest ~= track2.lowest ) then
		return track1.lowest < track2.lowest
	else
		return strcmputf8i(track1.name, track2.name) < 0;
	end
end

local tooltips = {}
local function getTooltip()
	local tooltip = nil;
	for i = 1, #tooltips do
		if ( tooltips[i]:GetOwner() == nil ) then
			tooltip = tooltips[i]
			break;
		end
	end
	if ( not tooltip ) then
		tooltip = CreateFrame("GameTooltip", "MissionMinderReportSummaryTooltip" .. (#tooltips+1), nil, "MissionMinderReportSummaryTooltip")
		DEBUG("Created %s", tooltip:GetName())
		tinsert(tooltips, tooltip)
	end

	tooltip:ClearAllPoints()
	return tooltip
end

local function hideReportTooltip(self)
	for i = 1, #tooltips do
		tooltips[i]:Hide()
	end
end

local function showReportTooltip(self)
	local now = time();

	local tracking = {}
	local characters = Lib.Globals.Characters
	for player, tbl in pairs(characters) do
		local lowest = nil;
		local completed = 0

		if ( tbl.missions ) then
			for missionID, mission in pairs(tbl.missions) do
				local endDate = mission.endDate
				if ( endDate <= now ) then
					lowest = 0;
					completed = completed + 1
				elseif ( not lowest or endDate - now < lowest ) then
					lowest = endDate - now;
				end
			end
		end

		if ( lowest ) then
			local tbl = {}
			tbl.name = player
			tbl.lowest = lowest
			tbl.completed = completed

			tinsert(tracking, tbl)
		end
	end

	local tooltips = { GameTooltip }

	-- Mission progress should always show when available
	if ( #tracking > 0 ) then
		local tooltip = getTooltip()
		tooltip:SetOwner(self, "ANCHOR_NONE")
		tooltip:SetPoint("TOPLEFT", tooltips[#tooltips], "BOTTOMLEFT")

		tooltip:SetText(L_GARRISON_REPORT_SUMMARY_MISSIONS, 1, 1, 1)
		tooltip:AddLine(L_GARRISON_REPORT_SUMMARY_MISSIONS_DESCRIPTION)
		tooltip:AddLine(" ")
		table.sort(tracking, orderTrackedMissions);
		for i = 1, #tracking do
			if ( tracking[i].completed > 0 ) then
				tooltip:AddDoubleLine(tracking[i].name, tracking[i].completed .. " " .. COMPLETED, 1, 1, 1, 0, 1, 0);
			else
				tooltip:AddDoubleLine(tracking[i].name, module.secondsToString(tracking[i].lowest), 1, 1, 1, 0.7, 0.7, 0.7);
			end
		end
		tinsert(tooltips, tooltip)
	end

	-- Building progress should always show when applicable
	local buildings = {}
	for player, tbl in pairs(characters) do
		if ( tbl.buildings ) then
			for _, building in pairs(tbl.buildings) do
				if ( building.isBuilding or building.canActivate ) then
					tinsert(buildings, {
						["player"] = player,
						["name"] = building.name,
						["isComplete"] = (now >= building.endDate),
						["endDate"] = building.endDate
					})
				end
			end
		end
	end
	if ( #buildings > 0 ) then
		local tooltip = getTooltip()
		tooltip:SetOwner(self, "ANCHOR_NONE")
		tooltip:SetPoint("TOPLEFT", tooltips[#tooltips], "BOTTOMLEFT")

		tooltip:SetText(L_GARRISON_REPORT_SUMMARY_BUIDLINGS, 1, 1, 1)
		tooltip:AddLine(L_GARRISON_REPORT_SUMMARY_BUIDLINGS_DESCRIPTION)

		table.sort(buildings, orderBuildings)
		local lastPlayer = nil
		for i = 1, #buildings do
			if ( lastPlayer ~= buildings[i].player ) then
				tooltip:AddLine(" ")
				tooltip:AddLine(buildings[i].player)
				lastPlayer = buildings[i].player
			end

			if ( buildings[i].isComplete ) then
				tooltip:AddDoubleLine(buildings[i].name, COMPLETED, 1, 1, 1, 0, 1, 0);
			else
				tooltip:AddDoubleLine(buildings[i].name, module.secondsToString(buildings[i].endDate - now), 1, 1, 1, 0.7, 0.7, 0.7)
			end
		end
		tinsert(tooltips, tooltip)
	end

	-- if any GarrisonCache is > 356 (less than 1 day from maxing)
	local garrisonCaches = {};
	for player, tbl in pairs(characters) do
		local cache = module.estimateGarrisonCache(player)
		if ( cache and cache >= GARRISON_CACHE_WARN_AMOUNT ) then
			tinsert(garrisonCaches, {
				["name"] = player,
				["amount"] = cache
			});
		end
	end
	if ( #garrisonCaches > 0 ) then
		local tooltip = getTooltip()
		tooltip:SetOwner(self, "ANCHOR_NONE")
		tooltip:SetPoint("TOPLEFT", tooltips[#tooltips], "BOTTOMLEFT")

		tooltip:SetText(L_GARRISON_REPORT_SUMMARY_CACHE, 1, 1, 1)
		tooltip:AddLine(L_GARRISON_REPORT_SUMMARY_CACHE_DESCRIPTION)
		tooltip:AddLine(" ")

		table.sort(garrisonCaches, orderGarrisonCache)
		for i = 1, #garrisonCaches do
			tooltip:AddDoubleLine(garrisonCaches[i].name, garrisonCaches[i].amount .. " " .. createInlineTexture(GARRISON_RESOURCES_TEXTURE), 1, 1, 1, 1, 0, 0);
		end
		tinsert(tooltips, tooltip)
	end

	-- if any work orders are nearly complete and need to be refilled in less than a day
	local workorders = {}
	for player, tbl in pairs(characters) do
		if ( tbl.workorders ) then
			for i = 1, #tbl.workorders do
				local workorder = tbl.workorders[i]
				local name = workorder.name
				local shipmentCapacity = workorder.shipmentCapacity
				local shipmentsReady = workorder.shipmentsReady
				local shipmentsTotal = workorder.shipmentsTotal
				local creationTime = workorder.creationTime
				local duration = workorder.duration

				if ( name and shipmentCapacity and shipmentCapacity > 0  and shipmentsTotal ) then
					if ( duration > 0 and now > creationTime + duration ) then -- Adjust ready based on elapsed time
						shipmentsReady = min(shipmentsTotal, shipmentsReady + (now - creationTime) / duration)
					end

					local timeRemaining = (shipmentsTotal - shipmentsReady) * duration
					if ( timeRemaining < 1 * DAYS ) then
						tinsert(workorders, {
							["player"] = player,
							["name"] = name,
							["ready"] = floor(shipmentsReady),
							["total"] = shipmentsTotal
						});
					end
				end
			end
		end
	end
	if ( #workorders > 0 ) then
		local tooltip = getTooltip()
		tooltip:SetOwner(self, "ANCHOR_NONE")
		tooltip:SetPoint("TOPLEFT", tooltips[#tooltips], "BOTTOMLEFT")

		tooltip:SetText(L_GARRISON_REPORT_SUMMARY_WORKORDERS, 1, 1, 1)
		tooltip:AddLine(L_GARRSION_REPORT_SUMMARY_WORKORDERS_DESCRIPTION)
		sort(workorders, orderWorkOrders)

		local lastPlayer = nil
		for i = 1, #workorders do
			if ( lastPlayer ~= workorders[i].player ) then
				tooltip:AddLine(" ")
				tooltip:AddLine(workorders[i].player)
				lastPlayer = workorders[i].player
			end

			tooltip:AddDoubleLine(workorders[i].name, workorders[i].ready .. "/" .. workorders[i].total, 1, 1, 1, 0, 1, 0);
		end
		tinsert(tooltips, tooltip)
	end


	-- Update tooltip width and right anchors
	local largestWidth = 0
	for i = 1, #tooltips do
		local tooltip = tooltips[i]
		tooltip:Show()
		largestWidth = max(largestWidth, tooltip:GetWidth())
	end
	for i = 1, #tooltips do
		local tooltip = tooltips[i]
		local difference = largestWidth - tooltip:GetWidth()
		tooltip:SetWidth(largestWidth)

		for line = 1, 100 do
			local left = _G[tooltip:GetName() .. "TextLeft" .. line]
			if (not left or not left:IsShown()) then -- No more work
				break;
			end

			local right = _G[tooltip:GetName() .. "TextRight" .. line]
			if ( right:IsShown() ) then
				local anchor, relobj, relanchor, x, y = right:GetPoint(1);
				right:SetPoint(anchor, relobj, relanchor, x + difference, y);
			end
		end
	end
end

local function trackGarrisonCache(event, quanity)
	Lib.Globals.Characters[CURRENT_PLAYER].GarrisonCache = time()
end

local function trackBuilding(event, plotID)
	if ( not plotID ) then
		return;
	end

	local buildings = Lib.Globals.Characters[CURRENT_PLAYER].buildings
	if ( not buildings ) then
		buildings = {}
		Lib.Globals.Characters[CURRENT_PLAYER].buildings = buildings
	end

	local workorders = Lib.Globals.Characters[CURRENT_PLAYER].workorders
	if ( not workorders ) then
		workorders = {}
		Lib.Globals.Characters[CURRENT_PLAYER].workorders = workorders
	end

	local id, name, texPrefix, icon, rank, isBuilding, timeStart, buildTime, canActivate, canUpgrade, isPrebuilt = C_Garrison.GetOwnedBuildingInfoAbbrev(plotID);
	if ( isBuilding or canActivate ) then -- building is constructing, add to workorders if not currently listed
		buildings[plotID] = {
			["plotID"] = plotID,
			["buildingID"] = id,
			["name"] = name,
			["texPrefix"] = texPrefix,
			["icon"] = icon,
			["rank"] = rank,
			["isBuilding"] = isBuilding,
			["timeStart"] = timeStart,
			["buildTime"] = buildTime,
			["endDate"] = timeStart + buildTime,
			["canActivate"] = canActivate,
			["canUpgrade"] = canUpgrade,
			["isPrebuilt"] = isPrebuilt
		}

		local workorder;
		for i = 1, #workorders do
			if (workorders[i].plotID == plotID ) then
				workorder = workorders[i]
				break;
			end
		end
		if ( not workorder ) then
			tinsert(workorders, {
				["texture"] = texture,
				["name"] = name,
				["buildingID"] = buildingID,
				["plotID"] = plotID
			});
		end
	else
		buildings[plotID] = nil
	end
end

local function trackShipments(event)
	local workorders = Lib.Globals.Characters[CURRENT_PLAYER].workorders
	if ( not workorders ) then
		workorders = {}
		Lib.Globals.Characters[CURRENT_PLAYER].workorders = workorders
	end

	local workorderIndex = 1;
	local buildings = C_Garrison.GetBuildings();
	for i = 1, #buildings do
		local plotID = buildings[i].plotID
		trackBuilding(nil, plotID)

		local buildingID = buildings[i].buildingID;
		if ( buildingID) then
			local name, texture, shipmentCapacity, shipmentsReady, shipmentsTotal, creationTime, duration, timeleftString, itemName, itemIcon, itemQuality, itemID = C_Garrison.GetLandingPageShipmentInfo(buildingID);

			if ( name and shipmentCapacity > 0 ) then
				local workorder = workorders[workorderIndex]
				if ( not workorder ) then
					workorder = {}
					workorders[workorderIndex] = workorder
				end
				workorderIndex = workorderIndex + 1;

				workorder.texture = texture
				workorder.name = name
				workorder.buildingID = buildingID
				workorder.plotID = plotID;
				workorder.creationTime = creationTime
				workorder.duration = duration
				workorder.itemName = itemName

				if ( workorder.shipmentCapacity ~= shipmentCapacity or workorder.shipmentsReady ~= shipmentsReady or workorder.shipmentsTotal ~= shipmentsTotal) then
					DEBUG("Updating building %d \"%s\" workorder: %d / %d (%d ready)", buildingID, name, shipmentsTotal, shipmentCapacity, shipmentsReady)
				end

				workorder.shipmentCapacity = shipmentCapacity
				workorder.shipmentsReady = shipmentsReady
				workorder.shipmentsTotal = shipmentsTotal
			end
		end
	end
	for i = workorderIndex, #workorders do
		workorders[i] = nil
	end
end

local function trackWorkOrder(event, plotID, maxShipments)
	local numPending = C_Garrison.GetNumPendingShipments();
	if (not numPending) then
		return;
	end

	if ( Lib.ptr and C_Garrison.IsOnShipmentQuestForNPC() ) then
		maxShipments = 1;
	end

	--local available = maxShipments - numPending;

	-- Update only the workorder for this specific building
	local buildingID, name = C_Garrison.GetOwnedBuildingInfoAbbrev(plotID);
	local workorders = Lib.Globals.Characters[CURRENT_PLAYER].workorders or {}
	for i = 1, #workorders do
		local workorder = workorders[i]
		if ( workorder.buildingID == buildingID and workorder.duration and workorder.duration > 0 ) then
			if ( workorder.shipmentCapacity ~= maxShipments or workorder.shipmentsTotal ~= numPending) then
				DEBUG("Updating building %d \"%s\" workorder: %d / %d", buildingID, name, numPending, maxShipments)
			end

			workorder.shipmentCapacity = maxShipments
			workorder.shipmentsTotal = numPending
			break;
		end
	end
end

local delayedRushOrderUpdate = nil
local function updateForRushOrder(event, spellID)
	if ( delayedRushOrderUpdate ) then
		delayedRushOrderUpdate:Cancel()
		delayedRushOrderUpdate = nil
	end

	delayedRushOrderUpdate = C_Timer.NewTimer(1 * SECONDS, C_Garrison.RequestLandingPageShipmentInfo);
end

local function initialize()
	DEBUG("   |cffffffffpreload-Tracker|r initialized");

	-- Display mission status on GarrisonReport tooltip
	GarrisonLandingPageMinimapButton:HookScript("OnEnter", showReportTooltip)
	GarrisonLandingPageMinimapButton:HookScript("OnLeave", hideReportTooltip)

	-- Track garrison cache across multiple characters
	Lib.RegisterCallback("tracker", "GARRISON_CACHE_LOOTED", trackGarrisonCache)

	-- Shipment info
	Lib.RegisterCallback("tracker", "SHIPMENT_UPDATE_AVAILABLE", trackShipments)

	-- Specific work order updated
	Lib.RegisterCallback("tracker", "WORKORDER_UPDATE_AVAILABLE", trackWorkOrder)

	-- Building info
	Lib.RegisterCallback("tracker", "BUILDING_UPDATE_AVAILABLE", trackBuilding)

	-- Handle rush order
	Lib.RegisterCallback("tracker", "RUSH_ORDER_USED", updateForRushOrder)
end

Lib:RegisterEvent("tracker", "ADDON_LOADED", initialize);
